package contrasts;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * This class represents a split (set1-set2).
 * 
 * A split contains two sets of populations. The goal of this class is to
 * compute a reliability score of the split for a given class. Therefore the
 * apply() method needs contrasts where a class is significantly different from
 * zero (CI does not cover zero). Then following score is computed:
 * 
 * <ul>
 * <li>{@code contrasts-between} = nonzero contrasts between set1 and set2</li>
 * <li>{@code all-contrasts-between} = all theoretical possible contrasts
 * between set1 and set2 ({@code set1.size * set2.size}).</li>
 * <li>{@code contrasts-within} = nonzero contrasts within set1 and within set2</li>
 * <li>{@code all-contrasts-within} = all theoretical possible contrasts within
 * set1 and within set2: {@code ((set1.size * (set1.size - 1.0)) / 2.0) + ((set2.size *
 * (set2.size - 1.0)) / 2.0}.</li>
 * </ul>
 * 
 * Now the score exists of the subtraction of the nonzero contrasts between
 * minus the nonzero contrasts within (use proportion for normalization):<br>
 * <br>
 * 
 * {@code score = contrasts-between/all-contrasts-between -
 * contrasts-within/all-contrasts-within}<br>
 * <br>
 * 
 * The score ranges from -1 to 1
 * <ul>
 * <li>-1: This means that for the class the split is totally unlikely, because
 * all contrasts are different within the sets but no contrast is different
 * between the sets.</li>
 * <li>-1 < score < 0: A negative value notes that the split is more unlikely
 * for that class.</li>
 * <li>0: Equal nonzero contrasts within and between the sets (normalized).</li>
 * <li>0 < score < 1: A positive value notes that the split is more likely for
 * that class</li>
 * <li>1:This means that for the class the split is as likely as possible,
 * because all contrasts are different between the sets but no contrast is
 * different within the sets.</li>
 * </ul>
 * 
 * @author <a href="mailto:max.schubach@informatik.uni-tuebingen.de">Max
 *         Schubach</a>
 * 
 * @version 1.1
 * 
 */
public class Split {

	/**
	 * Population set one of the Split. String array of populations.
	 */
	private String[] set1;
	/**
	 * Population set two of the Split. String array of populations.
	 */
	private String[] set2;

	/**
	 * A map which contains for every class the given split score.
	 */
	private Map<String, Double> splitScores;

	/**
	 * A map which contains for every class the average p-value.
	 */
	private Map<String, Double> pValueSplitScores;

	/**
	 * Default constructor of a Split
	 * 
	 * @param set1
	 *            population set 1
	 * @param set2
	 *            population set 2
	 */
	protected Split(String[] set1, String[] set2) {
		this.set1 = set1;
		this.set2 = set2;
	}

	/**
	 * @return A map which contains for every class the given split score.
	 */
	public Map<String, Double> getSplitScores() {
		// TODO Auto-generated method stub
		return this.splitScores;
	}

	/**
	 * @return A map which contains for every class the average p-value split
	 *         score.
	 */
	public Map<String, Double> getPValueSplitScores() {
		// TODO Auto-generated method stub
		return this.pValueSplitScores;
	}

	/**
	 * Apply classes that are significantly different from zero to the split and
	 * generate the split score.
	 * 
	 * @see Split More details about the split score
	 * 
	 * @param classesExcludeZero
	 *            A map of contrasts which contains class names with confidence
	 *            intervals significantly different from zero
	 */
	protected void apply(
			Map<String, Map<String, ArrayList<String>>> classesExcludeZero) {

		computeSplitScores(classesExcludeZero);

	}

	/**
	 * @param contrastsExcludeZero
	 * @param contrastsOfClasses
	 */
	public void apply(
			Map<String, Map<String, ArrayList<String>>> contrastsExcludeZero,
			Map<String, Map<String, Map<String, ConfidenceInterval>>> contrastsOfClasses) {
		apply(contrastsExcludeZero);
		computePValueScores(contrastsOfClasses);

	}

	/**
	 * @param contrastsOfClasses
	 */
	private void computePValueScores(
			Map<String, Map<String, Map<String, ConfidenceInterval>>> contrastsOfClasses) {
		this.pValueSplitScores = new HashMap<String, Double>();
		Map<String, Double> pValueSplitScoresWithin = new HashMap<String, Double>();
		Map<String, Double> pValueSplitScoresBetween = new HashMap<String, Double>();

		double norm_between = this.set1.length * this.set2.length;
		double norm_within = binom(this.set1.length) + binom(this.set2.length);

		// compute between
		pValueSplitScoresBetween = computePValueSplitScoresHelper(this.set1,
				this.set2, pValueSplitScoresBetween, contrastsOfClasses);
		// compute within
		pValueSplitScoresWithin = computePValueSplitScoresHelper(this.set1,
				this.set1, pValueSplitScoresWithin, contrastsOfClasses);
		pValueSplitScoresWithin = computePValueSplitScoresHelper(this.set2,
				this.set2, pValueSplitScoresWithin, contrastsOfClasses);

		Iterator<Map.Entry<String, Double>> entries = pValueSplitScoresBetween
				.entrySet().iterator();
		while (entries.hasNext()) {
			Map.Entry<String, Double> class_entry = entries.next();
			String class_name = class_entry.getKey();

			double score1 = class_entry.getValue() / norm_between;
			if (pValueSplitScoresWithin.containsKey(class_name)) {
				double score2 = pValueSplitScoresWithin.get(class_name)
						/ norm_within;
				this.pValueSplitScores.put(class_name,
						splitScoreLinear(score2,score1));
			} else {
				this.pValueSplitScores.put(class_name, score1);
			}
		}

		entries = pValueSplitScoresWithin.entrySet().iterator();
		while (entries.hasNext()) {
			Map.Entry<String, Double> class_entry = entries.next();
			String class_name = class_entry.getKey();
			if (!pValueSplitScoresBetween.containsKey(class_name)) {
				double score = 0.0 - class_entry.getValue() / norm_within;
				this.pValueSplitScores.put(class_name, score);
			}
		}

	}

	/**
	 * @param set1
	 * @param set2
	 * @param scores
	 * @param contrastsOfClasses
	 * @return
	 */
	private static Map<String, Double> computePValueSplitScoresHelper(
			String[] set1,
			String[] set2,
			Map<String, Double> scores,
			Map<String, Map<String, Map<String, ConfidenceInterval>>> contrastsOfClasses) {
		for (String dataset1 : set1) {
			for (String dataset2 : set2) {
				if (dataset1.equals(dataset2)) {
					continue;
				}
				if (contrastsOfClasses.containsKey(dataset1)
						&& contrastsOfClasses.get(dataset1).containsKey(
								dataset2)) {
					for (Map.Entry<String, ConfidenceInterval> class_entry : contrastsOfClasses
							.get(dataset1).get(dataset2).entrySet()) {
						String class_name = class_entry.getKey();
						ConfidenceInterval ci = class_entry.getValue();
						if (!scores.containsKey(class_name)) {
							scores.put(class_name, ci.pValue(0.0));
						} else {
							double actualValue = scores.get(class_name);
							scores.put(class_name, actualValue + ci.pValue(0.0));
						}
					}
				}
			}
		}
		if (set1 == set2) {
			for (Map.Entry<String, Double> scores_entry : scores.entrySet()) {
				scores.put(scores_entry.getKey(), scores_entry.getValue() / 2.0);
			}
		}
		return scores;
	}

	/**
	 * Helper method of apply()
	 * 
	 * @param classesExcludeZero
	 *            A map of contrasts which contains class names with confidence
	 *            intervals significantly different from zero
	 */
	private void computeSplitScores(
			Map<String, Map<String, ArrayList<String>>> classesExcludeZero) {
		Map<String, Double> splitValuesBetween = new HashMap<String, Double>();
		Map<String, Double> splitValuesWithin = new HashMap<String, Double>();
		this.splitScores = new HashMap<String, Double>();

		double norm_between = this.set1.length * this.set2.length;
		double norm_within = binom(this.set1.length) + binom(this.set2.length);

		// compute between
		splitValuesBetween = computeSplitScoresHelper(this.set1, this.set2,
				splitValuesBetween, classesExcludeZero);
		// compute within
		splitValuesWithin = computeSplitScoresHelper(this.set1, this.set1,
				splitValuesWithin, classesExcludeZero);
		splitValuesWithin = computeSplitScoresHelper(this.set2, this.set2,
				splitValuesWithin, classesExcludeZero);

		Iterator<Map.Entry<String, Double>> entries = splitValuesBetween
				.entrySet().iterator();
		while (entries.hasNext()) {
			Map.Entry<String, Double> class_entry = entries.next();
			String class_name = class_entry.getKey();

			double score1 = class_entry.getValue() / norm_between;
			if (splitValuesWithin.containsKey(class_name)) {
				double score2 = splitValuesWithin.get(class_name) / norm_within;
				this.splitScores.put(class_name,
						splitScoreLinear(score1, score2));
			} else {
				this.splitScores.put(class_name, score1);
			}
		}

		entries = splitValuesWithin.entrySet().iterator();
		while (entries.hasNext()) {
			Map.Entry<String, Double> class_entry = entries.next();
			String class_name = class_entry.getKey();
			if (!splitValuesBetween.containsKey(class_name)) {
				double score = 0.0 - class_entry.getValue() / norm_within;
				this.splitScores.put(class_name, score);
			}
		}
	}

	/**
	 * Compute a linear spilt score.
	 * 
	 * @param x
	 *            Value of between score
	 * @param y
	 *            Value of within score
	 * @return Simply substraction of x,y: x-y
	 */
	private static double splitScoreLinear(double x, double y) {
		return x - y;
	}

	/**
	 * 
	 * Helper to compute the split score. Computes all possible combination
	 * within or between (set1 = set2) the sets.
	 * 
	 * @param set1
	 *            Set1 of populations
	 * @param set2
	 *            Set two of population (=set1 => splits within are computed)
	 * @param Map
	 *            where to update the scores
	 * @param classesIncludeZero
	 *            Contrasts of classes that are significantly different from
	 *            zero.
	 * @return The (updated) split score. (update is possible, because in the
	 *         within step the method is applied two times, for set1 & for set2)
	 */
	private static Map<String, Double> computeSplitScoresHelper(String[] set1,
			String[] set2, Map<String, Double> scores,
			Map<String, Map<String, ArrayList<String>>> classesIncludeZero) {
		Map<String, Double> new_score = new HashMap<String, Double>();
		for (String dataset1 : set1) {
			for (String dataset2 : set2) {
				if (dataset1.equals(dataset2)) {
					continue;
				}
				if (classesIncludeZero.containsKey(dataset1)
						&& classesIncludeZero.get(dataset1).containsKey(
								dataset2)) {
					for (String class_name : classesIncludeZero.get(dataset1)
							.get(dataset2)) {
						if (!new_score.containsKey(class_name)) {
							new_score.put(class_name, 1.0);
						} else {
							double actualValue = new_score.get(class_name);
							new_score.put(class_name, actualValue + 1.0);
						}
					}
				}
			}
		}
		if (set1 == set2) {
			for (Map.Entry<String, Double> scores_entry : new_score.entrySet()) {
				new_score.put(scores_entry.getKey(), scores_entry.getValue() / 2.0);
				if (!scores.containsKey(scores_entry.getKey())) {
					scores.put(scores_entry.getKey(), new_score.get(scores_entry.getKey()));
				} else {
					double actualValue = scores.get(scores_entry.getKey());
					scores.put(scores_entry.getKey(), actualValue + new_score.get(scores_entry.getKey()));
				}
			}
		} else {
			return new_score;
		}
		
		return scores;
	}

	/**
	 * Compute the binomial coefficient n over 2
	 * 
	 * @param x
	 * @return Binomial coefficient x over 2
	 */
	private static double binom(double x) {
		return (x * (x - 1.0)) / 2.0;
	}

}
